#pragma once

#include <iostream>
#include <boost/shared_ptr.hpp>
#include "func.h"
#include "value.h"
#include "range.h"


typedef boost::shared_ptr<class Node> NodePtr;


//! Internal representation of expressions as syntax trees.
class Node {
    //     enum NodeType { NODE_VALUE, NODE_MONOP, NODE_BINOP, NODE_TEROP,
    //                     NODE_DREF, NODE_CREF };
    //     NodeType _ntyp;
 public:
    enum ReturnType { RET_VALUE, RET_RANGE, RET_LIST };

    // It seems that the base class must contain at least one virtual function.
    virtual ReturnType rtyp() const =0; 
    virtual string strg() const =0;
    virtual void input_type_check() const =0;
    virtual Value eval_val() const =0;
    virtual Range eval_range() const =0;
};

class NodeValue : public Node {
 private:
    Value value;
 public:
    NodeValue( int i ) : value(i) {};
    NodeValue( double d ) : value(d) {};
    ReturnType rtyp() const { return RET_VALUE; }; 
    string strg() const { return value.strg(); }
    void input_type_check() const {}; // ok since we have no input
    Value eval_val() const { return value; };
    Range eval_range() const { throw ProgramError( "unexpected request for range" ); };
};

class NodeMonop : public Node {
 protected:
    NodePtr nod0;
 public:
    NodeMonop( const NodePtr &nod0_in ) : nod0(nod0_in) {};
};

class NodeBinop : public Node {
 protected:
    NodePtr nod0;
    NodePtr nod1;
 public:
    NodeBinop( const NodePtr nod0_in, const NodePtr nod1_in ) : nod0(nod0_in), nod1(nod1_in) {};
};

class NodeNeg : public NodeMonop {
 public:
    NodeNeg( const NodePtr &nod0_in ) : NodeMonop( nod0_in ) {};
    ReturnType rtyp() const { return RET_VALUE; };
    string strg() const { return "-"+nod0->strg(); }
    void input_type_check() const {
        if( nod0->rtyp()!=RET_VALUE )
            throw BadInput( "unexpected return type in "+strg() );
    };
    Value eval_val() const { return -( nod0->eval_val() ); };
    Range eval_range() const { throw ProgramError( "unexpected request for range" ); };
};

class NodeParenthesis : public NodeMonop {
 public:
    NodeParenthesis( const NodePtr &nod0_in ) : NodeMonop( nod0_in ) {};
    ReturnType rtyp() const { return RET_VALUE; };
    string strg() const { return "("+nod0->strg()+")"; }
    void input_type_check() const {
        if( nod0->rtyp()!=RET_VALUE )
            throw BadInput( "unexpected return type in "+strg() );
    };
    Value eval_val() const { return nod0->eval_val(); };
    Range eval_range() const { throw ProgramError( "unexpected request for range" ); };
};

class NodeBinopVVV : public NodeBinop {
 public:
    NodeBinopVVV( const NodePtr &nod0_in, const NodePtr &nod1_in ) : NodeBinop( nod0_in, nod1_in ) {};
    ReturnType rtyp() const { return RET_VALUE; }; 
    void input_type_check() const {
        if( nod0->rtyp()!=RET_VALUE )
            throw BadInput( "unexpected return type of " + nod0->strg() + " in "+strg() );
        else if( nod1->rtyp()!=RET_VALUE )
            throw BadInput( "unexpected return type of " + nod1->strg() + " in "+strg() );
    }
    Range eval_range() const { throw ProgramError( "unexpected request for range" ); };
};

class NodeAdd : public NodeBinopVVV {
 public:
    NodeAdd( const NodePtr &nod0_in, const NodePtr &nod1_in ) : NodeBinopVVV( nod0_in, nod1_in ) {};
    string strg() const { return nod0->strg()+"+"+nod1->strg(); }
    Value eval_val() const { return nod0->eval_val() + nod1->eval_val(); };
};
    
class NodeSub : public NodeBinopVVV {
 public:
    NodeSub( const NodePtr &nod0_in, const NodePtr &nod1_in ) : NodeBinopVVV( nod0_in, nod1_in ) {};
    string strg() const { return nod0->strg()+"-"+nod1->strg(); }
    Value eval_val() const { return nod0->eval_val() - nod1->eval_val(); };
};
    
class NodeMul : public NodeBinopVVV {
 public:
    NodeMul( const NodePtr &nod0_in, const NodePtr &nod1_in ) : NodeBinopVVV( nod0_in, nod1_in ) {};
    string strg() const { return nod0->strg()+"*"+nod1->strg(); }
    Value eval_val() const { return nod0->eval_val() * nod1->eval_val(); };
};
    
class NodeDiv : public NodeBinopVVV {
 public:
    NodeDiv( const NodePtr &nod0_in, const NodePtr &nod1_in ) : NodeBinopVVV( nod0_in, nod1_in ) {};
    string strg() const { return nod0->strg()+"/"+nod1->strg(); }
    Value eval_val() const { return nod0->eval_val() / nod1->eval_val(); };
};

class NodeRange : public NodeBinop {
 public:
    NodeRange( const NodePtr &nod0_in, const NodePtr &nod1_in ) : NodeBinop( nod0_in, nod1_in ) {};
    ReturnType rtyp() const { return RET_RANGE; }; 
    string strg() const { return nod0->strg()+".."+nod1->strg(); }
    void input_type_check() const {
        if( nod0->rtyp()!=RET_VALUE )
            throw BadInput( "unexpected return type of " + nod0->strg() + " in "+strg() );
        else if( nod0->rtyp()!=RET_VALUE )
            throw BadInput( "unexpected return type of " + nod1->strg() + " in "+strg() );
    }
    Range eval_range() const { return Range( nod0->eval_val(), nod1->eval_val() ); };
    Value eval_val() const { throw ProgramError( "unexpected request for value" ); };
};

class NodeInside : public NodeBinop {
 public:
    NodeInside( const NodePtr &nod0_in, const NodePtr &nod1_in ) : NodeBinop( nod0_in, nod1_in ) {};
    string strg() const { return nod0->strg()+" in "+nod1->strg(); }
    ReturnType rtyp() const { return RET_VALUE; }; 
    void input_type_check() const {
        if( nod0->rtyp()!=RET_VALUE )
            throw BadInput( "unexpected return type of " + nod0->strg() + " in "+strg() );
        else if( nod1->rtyp()!=RET_RANGE )
            throw BadInput( "unexpected return type of " + nod1->strg() + " in "+strg() );
    }
    Value eval_val() const { return Value( (int) nod1->eval_range().contains( nod0->eval_val() ) ); };
    Range eval_range() const { throw ProgramError( "unexpected request for range" ); };
};

  
class CTree {
 private:
    enum TreeTyp { _NOTREE=0, _MONOP, _BINOP, _TEROP,
                   _VAL, _REF, _FRF };
    TreeTyp typ;
    double val;
    CFunc* fun;
    CTree* arg1;
    CTree* arg2;
    CTree* arg3;
//    CRef*  ref;

 public:
    inline CTree() { 
        typ=_NOTREE; };
    inline CTree(double v_in) { typ=_VAL; val = v_in; };
    inline CTree(CFunc *f_in, CTree *a1) {
        if (f_in->narg!=1) {
            cout << "PROG ERR: CTree(1) with conflicting #arg\n";
            typ=_NOTREE;
            return;
        }
        typ=_MONOP; fun=f_in; arg1 = a1; 
    }
    inline CTree(CFunc *f_in, CTree *a1, CTree *a2) { 
        if (f_in->narg!=2) {
            cout << "PROG ERR: CTree(2) with conflicting #arg\n";
            typ=_NOTREE;
            return;
        }
        typ=_BINOP; fun=f_in; arg1 = a1; arg2 = a2; };
    inline CTree(CFunc *f_in, CTree *a1, CTree *a2, CTree *a3) { 
        if (f_in->narg!=3) {
            cout << "PROG ERR: CTree(3) with conflicting #arg\n";
            typ=_NOTREE;
            return;
        }
        typ=_TEROP; fun=f_in; arg1 = a1; arg2 = a2; arg3 = a3; };
//    inline CTree(CRef* ref_in) { 
//        typ=_REF; ref=ref_in; };
//    inline CTree(CRef* ref_in, CTree* a1) { 
//        typ=_FRF; ref=ref_in; arg1 = a1; };
    inline ~CTree() {
        switch(typ) {
        case _TEROP:
            delete arg3;
        case _BINOP:
            delete arg2;
        case _MONOP:
            delete arg1;
            break;
//        case _REF:
//            delete ref;
        default:
            ;
        }
    };

    CTree& operator=(const CTree&);
    double tree_curve_val(double x, vector<double>& P); // for curve fitting
    int tree_val(double *ret, class COlo* f=0, 
		 uint k=(uint)-1, uint j=(uint)-1, uint i=(uint)-1);
    int coord(CCoord *ret, COlo *f=0, uint k=(uint)-1);
//    uint npar(CRef::TVar r);
    string info(void);

 private:
//    void nparx(CRef::TVar v, uint *np);
};
