//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* olm.cpp: online memory                                                 *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <algorithm>
using namespace std;

#include "mystd.h"
#include "olm.h"
#include "asi.h"

//***************************************************************************//
//* class COld                                                              *//
//* online data files                                                       *//
//***************************************************************************//


COld::COld( class COlc const* c )
{
    name = c->name;
    lDoc = c->lDoc;
    xco = c->xco;
    yco = c->yco;
    ZCo = c->ZCo;
    RPar = c->RPar;
}

uint COld::nPts( uint j ) const {
    if (!nScan() || j>=nScan())
        throw string( "PROGRAM ERROR: nPts called with invalid j="+strg(j) );
    return VS[j].size();
}


//! Return number of points if it is the same for all scans. Else return 0.

uint COld::nPts() const {
    if (!nScan())
        throw string( "PROGRAM ERROR: nPts(void) called while nScan=0" );
    uint np = VS[0].size();
    for (uint j=0; j<nScan(); ++j)
        if (VS[j].size()!=np) return 0;
    return np;
}


//***************************************************************************//
//* namespace NOlm                                                          *//
//***************************************************************************//

namespace NOlm { // on-line memory (in olm.h predefined as extern)
    // the data:
    vector<class CEle> MOM;

    // the current selection:
    class CList FSel;
    class CList FSelOld;
    bool overwrite;
};


//***************************************************************************//
//* NOlm: basic administration of online memory                             *//
//***************************************************************************//

//! Clear file selection.
//  To be called after 'eiter' and before creation of new files.

void NOlm::SelNew(void)
{
    FSelOld = FSel;
    FSel = CList();
}

//! Go back to previous file selection.

void NOlm::SelPop(void)
{
    FSel = FSelOld;
    FSelOld = CList();
}

//! Append copy of file *olo to online memory.

CEle* NOlm::OloAdd( const COlo *olo )
{
    CEle e;
    if        ( const COld* olf = dynamic_cast<const COld*>(olo) ) {
        COld *p = new COld;
        *p = *olf;
        e = CEle( p );
    } else if ( const COlc* olc = dynamic_cast<const COlc*>(olo) ) {
        COlc *p = new COlc;
        *p = *olc;
        e = CEle( p );
    } else
        throw string("PROGAMM ERROR: invalid on-line object in OloDup");
    e.P()->as_on_disk = false;
    MOM.push_back( e );
    FSel.append( MOM.size()-1 );
    return &( MOM.back() );
}

//! Remove selected files from online memory.

void NOlm::OlfDel()
{
    for (uint i = FSel.size()-1; i!=(uint)-1; --i) {
        MOM[FSel.V[i]].predelete();
        MOM.erase(MOM.begin()+FSel.V[i]);
    }
    SelPop();
    SelPop();
    SelNew();
    if( MOM.size() > 0 )
        FSel = CList(0,MOM.size()-1);
    else
        FSel = CList();
}

//! Append copy of selected files to online memory.

void NOlm::OlfCopy()
{
    CEle *e;
    IterateEle eiter;
    SelNew();
    while ((e=eiter()))
        OloAdd(e->P());
}

//! Move files to end of online memory.

void NOlm::OlfExchange()
{
    CEle *e;
    IterateEle eiter;
    FSelOld = FSel;
    FSel = CList();
    while ((e=eiter()))
        OloAdd(e->P());
    for (uint i = FSelOld.size()-1; i!=(uint)-1; --i) {
        MOM[FSelOld.V[i]].predelete();
        MOM.erase(MOM.begin()+FSelOld.V[i]);
    }
    for ( uint i=0; i<FSelOld.size(); ++i )
        FSel.V[i] -= FSelOld.size();
}

//! For each file in selection, set fno, fin, fout, and call exec(...).

void NOlm::OlfLoop(void(*exec)(int fno, COld *fin, COld *fout))
{
    IterateD fiter;
    if (!overwrite) SelNew();
    COld *fin, ftmp, *fout;
    int fno;
    while((fin=fiter())) {
        fno = fiter.SelNo();
        cout << "-> operate on file " << fno << "\n";
        OUT_TO_IN( overwrite, fin, ftmp, fout );

        exec(fno, fin, fout);

        if(!(overwrite)) OloAdd(&ftmp);
    }
}


//***************************************************************************//
//* NOlm: scan selection                                                    *//
//***************************************************************************//

void NOlm::JSelAsk( string quest, string *jSel, CList *jLis )
{
    if ( FSel.size()<=0 )
        throw string( "query scan selection: no files to operate on" );

    IterateEle iter;
    CEle *e;
    uint nJmax = 0;
    while ( (e=iter()) )
        nJmax = max( nJmax, e->nScan() );

    if ( nJmax==1 )
        *jLis = CList("0");
    else {
        *jSel = wask( quest, *jSel );
        *jLis = CList ( *jSel, 0, nJmax-1 );
    }
}
