//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* opr.cpp: operate on data files                                         *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <boost/format.hpp>
using boost::format;
using namespace std;

#include "mystd.h"
#include "olm.h"
#include "expr.h"
#include "asi.h"
#include "opr.h"
#include "xax_lex.h"


//***************************************************************************//
//* operate via generic expressions                                         *//
//***************************************************************************//

//! Display data. Use boolean expression to display only a subset of the data.

void NOperate::Show( string subcmd ) 
{
    static string expr;
    expr = sask( "Show if", expr);
    if (expr=="") expr = "1";
    PTree T;
    user_xaxparse( expr.c_str(), &T );

    NOlm::IterateD fiter;
    const COld *fin;
    COlo *f;
    double lval;
    uint k, ni;
    string outj, outp;

    while((fin=fiter())) {
        k = fiter.SelNo();
        if ( fiter.size()>1 )
            printf( ".. file %d:\n", fiter.SelNo() );

        f = (COlo*) fin;

        // determine length of longest spectrum:
        ni = 0;
        for ( uint j=0; j<fin->nScan(); j++)
            ni = max( ni, fin->VS[j].size() );

        // printout by i:
        for ( uint i=0; i<ni; i++) {

            outj = "";
            outp = "";

            for( uint j=0; j<fin->nScan(); j++ ){
                const CScan *s;
                s = &(fin->VS[j]);
                if( i>= s->size() ) continue;
                T->tree_point_val( &lval, k, j, i );
                if( !lval ) continue;
                outj += strg(j) + " ";
                if      ( subcmd=="xyd" )
                    if( s->dy.size() )
                        outp += str( format( "%- 12.5g %- 12.5g %- 12.5g" )
                                     % s->x[i] % s->y[i] % s->dy[i] );
                    else
                        outp += str( format( "%- 12.5g %- 12.5g" )
                                     % s->x[i] % s->y[i] );
                else if ( subcmd=="y" )
                    outp += str( format( "%- 12.5g" ) % s->y[i] );
                else
                    throw "invalid subcmd " + subcmd;
                outp += " ";
            }

            if( outp=="" ) continue;

            printf( "%5u [%s] %s\n", i, outj.substr(0,outj.size()-1).c_str(), 
                    outp.substr(0,outp.size()-1).c_str() );
        }
    }
}


//! Delete or retain data points according to a boolean expression.

void NOperate::Select( bool askdelete )
{
    static string expr;
    expr = sask((askdelete ? "Delete" : "Retain"), expr);
    if (expr=="") return;
    PTree T;
    user_xaxparse( expr.c_str(), &T );

    NOlm::IterateEle eiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    const CEle *ein;
    const COlo *fin;
    COlo *ftmp;
    COld d;
    COlc c;
    const CScan *Sin;
    CScan Sout;
    uint k;

    while((ein=eiter())) {
        k = eiter.SelNo();
        printf( ".. operate on file %d\n", k );

        ftmp = ein->isD() ? (COlo*) &d : (COlo*) &c;

        if( ein->isD() ){
            d = *(ein->D());
            d.VS.clear();
        } else {
            c = *(ein->C());
            c.VC.clear();
        }            
        fin = ein->P();
        
        ftmp->lDoc.push_back((askdelete ? "md " : "mr ") + expr);
        
        for (uint j=0; j<ein->nScan(); j++) {
            // printf("DEBUG .. operate on scan %d\n", j);
            
            if( ein->isD() ){
                Sin = &(((COld*)fin)->VS[j]);
                Sout = *Sin;    // copy z stuff
                Sout.x.clear();
                Sout.y.clear();
                vector<double> lvec( Sin->size() );
                T->tree_vec_val( &lvec, k, j );
                for( uint i=0; i<Sin->size(); i++ )
                    if ( askdelete ^ (bool) lvec[i] )
                        Sout.push_xy(Sin->x[i], Sin->y[i]);
                if (Sout.size())
                    d.VS.push_back(Sout);
            } else {
                double lval;
                T->tree_point_val( &lval, k, j );
                if ( askdelete ^ (bool) lval )
                    c.VC.push_back(((COlc*)fin)->VC[j]);
            }                
        }
        if ( ein->isD() ? d.nScan() : c.nScan() )
            NOlm::OloAdd(ftmp);
        else 
            cout << "no data point left\n";
    }
}


//! Pointwise operation on x, y, or z<n>, according to arithmetic expression.

void NOperate::Pointwise( string llabel ) 
{
    CRef lref(llabel);
    if ( !lref.defined() )
        throw "invalid left-hand side " + llabel;

    string expr = sask("Function ?");
    if (expr=="") return;
    PTree T;
    user_xaxparse( expr.c_str(), &T );
 
    NOlm::IterateEle iter;
    if (!NOlm::overwrite) NOlm::SelNew();
    CEle *ein;
    COlo *fout;
    vector<double> vout;
    double dout;
    uint n;
    while( (ein=iter()) ){
        uint k = iter.SelNo();
        if (NOlm::overwrite) {
            fout = ein->P();
        } else {
            if( ein->isD() )
                fout = (COlo*) new COld( *((COld*) ein->P()) );
            else
                fout = (COlo*) new COlc( *((COlc*) ein->P()) );
        }
        fout->lDoc.push_back( "o" + lref.ref_info() + " " + expr);
        
        if      (lref.var==CRef::_X) 
            T->coord(&(fout->xco), k);
        else if (lref.var==CRef::_Y) 
            T->coord(&(fout->yco), k);
        else if (lref.var==CRef::_Z) {
            if( lref.num>= fout->ZCo.size() ){
                delete fout;
                throw string( "no such z coordinate to operate onto" );
            }
            T->coord(&(fout->ZCo[lref.num]), k);
        }

        for( uint j=0; j<ein->nScan(); j++ ){
            
            if (lref.pointwise()) {
                if( !ein->isD() ){
                    delete fout;
                    throw string( "no pointwise operation on curve" );
                }
                n = ((COld*)(ein->D()))->VS[j].size();
                vout.resize(n);
                T->tree_vec_val( &vout, k, j );
                if (lref.var==CRef::_X)
                    ((COld*)fout)->VS[j].x = vout;
                else
                    ((COld*)fout)->VS[j].y = vout;
            } else if (lref.var==CRef::_Z) {
                T->tree_point_val( &dout, k, j );
                if( ein->isD() )
                    ((COld*)fout)->VS[j].z[lref.num] = dout;
                else
                    ((COlc*)fout)->VC[j].z[lref.num] = dout;
            } else {
                delete fout;
                throw string( "PROGRAM ERROR/ invalid oxy" );
            }
			
        }
        if( !(NOlm::overwrite) ){
            NOlm::OloAdd( fout );
            delete fout;
        }
        
    }
}


//! Add new z coordinate, and set values according to arithmetic expression.

void NOperate::NewZ() 
{
    string expr = sask("Function ?");
    if (expr=="") return;
    PTree T;
    user_xaxparse( expr.c_str(), &T );

    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *fin, ftmp, *fout;
    COlo *f;
    double dout;
    uint k;
    while( (fin=fiter()) ) {
        k = fiter.SelNo();
        OUT_TO_IN( NOlm::overwrite, fin, ftmp, fout );
        f = (COlo*) fin;

        fout->lDoc.push_back("oz+ " +expr);
        fout->ZCo.push_back(CCoord());
        CRef lref(string("z")+strg(fout->ZCo.size()-1));
        T->coord(&(fout->ZCo[lref.num]), k);

        for (uint j=0; j<fin->nScan(); j++) {
            T->tree_point_val( &dout, k, j );
            fout->VS[j].z.push_back(dout);
        }
        if(!(NOlm::overwrite)) NOlm::OloAdd(&ftmp);
    }
}


//! Integral operation, using arithmetic expression.

void NOperate::Integral(void) 
{
    string expr = sask("Functional ?");
    if (expr=="") return;
    PTree T;
    user_xaxparse( expr.c_str(), &T );

    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *fin, fout;
    COlo *f;
    CScan S;
    CCoord zco;
    double xval, zval, result;
    bool savable;
    uint k, nz;
    while((fin=fiter())) {
        k = fiter.SelNo();
        nz = fin->ZCo.size();
        savable = nz > 0;

        fout = *fin;
        f = (COlo*) fin;
        fout.VS.clear();
        if (savable) {
            fout.lDoc.push_back("oi " +expr);

            zco = fin->ZCo.back();
            fout.ZCo.pop_back();
            fout.xco = zco;

            T->coord(&(fout.yco), k);
        }

        S.Clear();
        S.z = fin->VS[0].z;
        S.z.pop_back();
        for (uint j=0; j<fin->nScan(); j++) {
            T->tree_point_val( &result, k, j );
            if (!savable) {
                cout << "results: " << result << "\n";
                break;
            }
            xval = fin->VS[j].z[nz-1];
            S.push_xy(xval, result);
            if (nz>=2) { // new scan if jump in other z values
                zval = fin->VS[j].z[nz-2];
                if ((j+1)<fin->VS.size() && 
                    fin->VS[j+1].z[nz-2]!=zval) {
                    fout.VS.push_back(S);
                    S.Clear();
                    S.z = fin->VS[j+1].z;
                    S.z.pop_back();
                }
            }
        }
        if (savable) {
            fout.VS.push_back(S);
            NOlm::OloAdd(&fout);
        }
    }
};


//! Integral operations to generate not only new y, but also new x.

void NOperate::IntXY(void) 
{
    static int icolx = 0, icoly = 1;
    static CCoord xco, yco;
    icolx = iask( "x from column", icolx );
    if( icolx<0) return;
    icoly = iask( "y from column", icoly );
    if( icoly<0) return;
    xco = CCoord( "x", "" ); // .Ask( "x coordinate", ASK_DEF_GIVEN, xco );
    yco = CCoord( "y", "" ); // .Ask( "y coordinate", ASK_DEF_GIVEN, yco );

    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *fin, fout;
    COlo *f;
    CScan S;
    CCoord zco;
    double zval;
    uint k, nz, n;
    while((fin=fiter())) {
        k = fiter.SelNo();
        nz = fin->ZCo.size();
        if( nz<1 )
            throw "no z in file " + strg(fiter.SelNo());
        fout = *fin;
        f = (COlo*) fin;
        fout.VS.clear();
        fout.lDoc.push_back( "oixy " + strg(icolx) + " " + strg(icoly) );
        fout.ZCo.pop_back();
        fout.xco = xco;
        fout.yco = yco;
        S.Clear();
        if( !(fin->nScan()) )
            throw string( "input file has no scans" );
        S.z = fin->VS[0].z;
        S.z.pop_back();
        for (uint j=0; j<fin->nScan(); j++) {
            n = fin->VS[j].size();
            if( icolx>=n || icoly>=n )
                throw "not enough columns in scan " + strg(j);
            S.push_xy( fin->VS[j].y[icolx], fin->VS[j].y[icoly] );
            if (nz>=2) { // new scan if jump in other z values
                zval = fin->VS[j].z[nz-2];
                if ((j+1)<fin->VS.size() && 
                    fin->VS[j+1].z[nz-2]!=zval) {
                    fout.VS.push_back(S);
                    S.Clear();
                    S.z = fin->VS[j+1].z;
                    S.z.pop_back();
                }
            }
        }
        fout.VS.push_back(S);
        NOlm::OloAdd(&fout);
    }
};


//! Determine eigenvalue and eigenvectors of quadratic data set.

#include<gsl/gsl_eigen.h>

void NOperate::Eigen() 
{
    NOlm::IterateD fiter;
    NOlm::SelNew();
    COld *f0, fout0, fout1;
    CScan Sout0;
    uint i, n;
    gsl_vector *EVal;
    gsl_matrix *Data, *EVec;
    gsl_eigen_symmv_workspace *Work;

    while((f0=fiter())) {
        if( !( n = f0->nPts() ) ){
            printf( "! Eigen: scans must have equal length\n" );
            continue;
        } else if ( n != f0->nScan() ) {
            printf( "! Eigen: file must contain a quadratic matrix\n" );
            continue;
        }

        fout0 = *f0;
        fout1 = *f0;

        fout0.lDoc.push_back( "y := eigenvalues" );
        fout1.lDoc.push_back( "y := eigenvectors" );

        fout0.xco = CCoord("idx", "");
        fout0.yco = CCoord("lambda", "");
        fout0.ZCo.clear();
        fout0.VS.clear();
        fout1.ZCo.clear();
        fout1.ZCo.push_back( CCoord("idx", "") );
            
        EVal = gsl_vector_alloc( n );
        Data = gsl_matrix_alloc( n, n );
        EVec = gsl_matrix_alloc( n, n );
        Work = gsl_eigen_symmv_alloc( n );

        for( uint j=0; j<n; j++ ){
            for( i=0; i<n; i++ ){
                gsl_matrix_set( Data, i, j, f0->VS[j].y[i] );
            }
        }

        gsl_eigen_symmv( Data, EVal, EVec, Work );
        gsl_eigen_symmv_sort( EVal, EVec, GSL_EIGEN_SORT_VAL_DESC  );
        Sout0.Clear();
        for( i=0; i<n; ++i ){
            Sout0.push_xy( (double) i, EVal->data[i] );
            printf( "eval %d: %g\n", i, EVal->data[i] );
        }
        fout0.VS.push_back( Sout0 );

        for( uint j=0; j<n; j++ ){
            fout1.VS[j].Clear();
            fout1.VS[j].z.push_back( (double) j);
            for( i=0; i<n; i++ ){
                fout1.VS[j].push_xy( (double) i,
                                       gsl_matrix_get( EVec, i, j ) );
            }
        }

        gsl_vector_free( EVal );
        gsl_matrix_free( Data );
        gsl_matrix_free( EVec );
        gsl_eigen_symmv_free( Work );

        NOlm::OloAdd(&fout0);
        NOlm::OloAdd(&fout1);
    }
};


//! Calculate correlation coefficients.

void NOperate::CalcCorrel() 
{
    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *f0, *f1, fout;
    CScan *S0, *S1, S;
    double s0, s1, s00, s11, s01, y0, y1, cc;
    static int k1=-1;
    uint i, j, jj, n, neff, nmin, nmax, ndim;
    static int modeIgnore=0, modeFile=0;
    static double yIgnore=0.;
    bool j2floating;
    cout << "Calculate correlation coefficients between:\n"
        "  (1) x[j] and y[j]\n"
        "  (2) y[j] and y'[j]\n"
        "  (3) y[j] and y[j']\n"
        "  (4) y[j] and y'[j']\n"; 
    modeFile = iask( "Option", modeFile );
    if( modeFile<=0 ) return;

    modeIgnore = iask( "Ignore (0:nothing 1:equal yI 2: below yI",
                       modeIgnore );
    if( modeIgnore !=0 ){
        yIgnore = dask( "yI (ignore when equal/below)", yIgnore );
    }
    
    while((f0=fiter())) {
        
        if( modeFile==2 || modeFile==4 ){
            k1 = k1<0 ? iask( "second file" ) : iask( "second file", k1 );
            if( k1<0 )
                return;
            if( k1>=NOlm::MOM.size() ){
                printf( "invalid file number %i\n", k1 );
                return;
            }
            if( !(f1=NOlm::MOM[k1].D()) ){
                printf("! is not a data file\n");
                return;
            }
        } else {
            f1 = f0;
        }
        
        j2floating = modeFile==3 || modeFile==4;
        // dimenionality of output:
        if     ( j2floating && f1->nScan()>1 )
            ndim = 2;
        else if( f0->nScan()>1 )
            ndim = 1;
        else
            ndim = 0;
        
        fout = *f0;

        fout.lDoc.push_back( "y := correlation with file "+strg(k1) );
        fout.yco = CCoord("cor["+f0->yco.name+"]", "");
        fout.ZCo.clear();
        if       ( ndim==2 ){
            fout.ZCo.push_back( f0->ZCo[0] );
            fout.xco = f0->ZCo[0];
        } else if( ndim==1 ){
            fout.xco = f0->ZCo[0];
            fout.VS.clear();
            fout.VS.push_back( CScan() );
        } else {
            // do not save
        }

            
        for( j=0; j<f0->nScan(); j++ ){

            S0 = &(f0->VS[j]);
            n = S0->size();
            nmin = n;
            nmax = 0;

            if( ndim==2 ){
                fout.VS[j].Clear();
                fout.VS[j].z.push_back( S0->z[0] );
            }

            for( jj = (j2floating ? 0 : j);
                 jj < (j2floating ? f1->nScan() : j+1); ++jj ){

                S1 = &(f1->VS[jj]);

                if( S1->size()!=n ){
                    printf( "S0[j=%i].size=%i not equal S1[jj=%i].size=%u\n",
                            j, n, jj, S1->size() );
                    return;
                }

                s0 = 0; s1 = 0; s00 = 0; s11 = 0; s01 = 0; neff = 0;
                for (i=0; i<n; i++) {
                    y0 = S0->y[i];
                    y1 = modeFile==1 ? S1->x[i] : S1->y[i];
                    if( ( modeIgnore==1 && (y0==yIgnore || y1==yIgnore ) ) ||
                        ( modeIgnore==2 && (y0< yIgnore || y1< yIgnore ) ) )
                        continue;
                    neff += 1;
                    s0 += y0;
                    s1 += y1;
                    s00 += SQR( y0 );
                    s11 += SQR( y1 );
                    s01 += y0 * y1;
                }
                if( neff<nmin ) nmin = neff;
                if( neff>nmax ) nmax = neff;
                cc = neff>0 ? (s01-s0*s1/neff) /
                    sqrt((s00-s0*s0/neff)*(s11-s1*s1/neff)) : 0;

                if        ( ndim==2 ){
                    fout.VS[j].push_xy( S1->z[0], cc);
                } else if ( ndim==1 ) {
                    fout.VS[0].push_xy( S0->z[0], cc);
                } else {
                    printf( "1x1: corr = %g\n", cc );
                }
            }
            printf( "column %d: min %d max %d nonnil correlations\n",
                    j, nmin, nmax );
        }
        if( ndim>0 ) NOlm::OloAdd(&fout);
}
};


//! Calculate WHAT???

void NOperate::CalcCorrelDiff() 
{
    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *fin, fout;
    CScan *S0, *S1, S;
    double s0, s1, s00, s11, s01, y0, y1;
    uint i, n, neff;
    printf(" warning: PISA version: ignore 0.0 data\n" );
    while((fin=fiter())) {
        if( !( n = fin->nPts() ) ){
            printf( "! CalcCorrelDiff: spectra must have equal length\n" );
            printf( "(do not delete invalid items: set them to zero)\n" );
            continue;
        }
        fout = *fin;

        fout.lDoc.push_back( "y := impact on correlation coefficient " );
        fout.yco = CCoord("dr^2["+fin->yco.name+"]", "");
        fout.ZCo.push_back( CCoord( fin->ZCo[0].name+"'", fin->ZCo[0].unit ) );
        fout.ZCo.push_back( CCoord("r^2["+fin->yco.name+"]", "") );
        fout.VS.clear();

        for (uint j0=0; j0<fin->nScan(); j0++) {
            for (uint j1=0; j1<fin->nScan(); j1++) {
                S0 = &(fin->VS[j0]);
                S1 = &(fin->VS[j1]);

                S = *S0;

                s0 = 0; s1 = 0; s00 = 0; s11 = 0; s01 = 0; neff = 0;
                for (i=0; i<n; i++) {
                    y0 = S0->y[i];
                    y1 = S1->y[i];
                    if( y0<0 || y1<0 ) continue;
                    neff += 1;
                    s0 += y0;
                    s1 += y1;
                    s00 += SQR( y0 );
                    s11 += SQR( y1 );
                    s01 += y0 * y1;
                }
                
                S.z.push_back( S1->z[0] );
                S.z.push_back( (s01-s0*s1/neff)/sqrt(
                                   (s00-s0*s0/neff)*(s11-s1*s1/neff)) );

                for (i=0; i<n; i++) {
                    y0 = S0->y[i];
                    y1 = S1->y[i];
                    if( y0<0 || y1<0 ){
                        S.y[i] = -1;
                    } else {
                        S.y[i] = ( (s01-y0*y1) - (s0-y0)*(s1-y1)/(neff-1) ) / 
                            sqrt(
                                ( (s00-y0*y0) - (s0-y0)*(s0-y0)/(neff-1) ) *
                                ( (s11-y1*y1) - (s1-y1)*(s1-y1)/(neff-1) ) );
                    }
                }
                
                fout.VS.push_back(S);
            }
        }
        NOlm::OloAdd(&fout);
    }
};


//! Calculate WHAT???

void NOperate::CalcAvgeDiff()
// very specifically for PISA
{
    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *f0, *f1, fout;
    CScan *S0, *S1, S;
    double s0, s1, y0, y1;
    uint i, n, neff;

    while((f0=fiter())) {
        if( !( n = f0->nPts() ) ){
            printf( "! CalcCorrel: scans must have equal length\n" );
            continue;
        }
        if( f1->nScan() <= 1 ){
            printf( "! Expect at least 2 scans\n" );
            return;
        }

        f1 = f0;
        fout = *f0;

        fout.lDoc.push_back( "y := average differences" );
        fout.yco = CCoord("<y>-<y'>", "");
        fout.ZCo.clear();
        fout.ZCo.push_back( f0->ZCo[0] );
        fout.xco = f0->ZCo[0];
            
        for (uint j=0; j<f0->nScan(); j++) {
            fout.VS[j].Clear();
            S0 = &(f0->VS[j]);
            fout.VS[j].z.push_back( S0->z[0] );
            for (uint jj=0; jj<f1->nScan(); ++jj) {
                S1 = &(f1->VS[jj]);
                neff = 0;
                s0 = 0;
                s1 = 0;
                for (i=0; i<n; i++) {
                    y0 = S0->y[i];
                    y1 = S1->y[i];
                    if( y0<0 || y1<0 )
                        continue;
                    neff += 1;
                    s0 += y0;
                    s1 += y1;
                }
                fout.VS[j].push_xy( S1->z[0], (s0-s1)/neff);
            }
        }
        NOlm::OloAdd(&fout);
    }
};


//***************************************************************************//
//* functional operations                                                   *//
//***************************************************************************//

void NOperate::FunctionalIntegrate()
{
    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *fin, ftmp, *fout;
    while((fin=fiter())) {
        OUT_TO_IN( NOlm::overwrite, fin, ftmp, fout );
        fout->yco.name = "Int d" + fin->xco.name + " " + fin->yco.name;
        fout->yco.unit = fin->xco.unit + " * " + fin->yco.unit;
        fout->lDoc.push_back( "ofi" );
        vector<double> *xin, *yin, *yout;
        for (uint j=0; j<fin->VS.size(); j++) {
            xin = &(fin->VS[j].x);
            yin = &(fin->VS[j].y);
            yout = &(fout->VS[j].y);
            uint n = yin->size();

            (*yout)[0] = 0;
            for (uint i=1; i<n; ++i) {
                (*yout)[i] = (*yout)[i-1] + 
                    ((*xin)[i] - (*xin)[i-1]) * 
                    ((*yin)[i] + (*yin)[i-1]) / 2;
            }

            fout->VS[j].x = *xin;
        }
        if(!(NOlm::overwrite)) NOlm::OloAdd(&ftmp);
    }
}

void NOperate::FunctionalDerivate()
{
    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *fin, ftmp, *fout;
    while((fin=fiter())) {
        OUT_TO_IN( NOlm::overwrite, fin, ftmp, fout );
        fout->yco.name = "d" + fin->yco.name + " / d" + fin->xco.name;
        fout->yco.unit = fin->yco.unit + " / " + fin->xco.unit;
        fout->lDoc.push_back( "ofd" );
        vector<double> *xin, *yin;
        double dx;
        for (uint j=0; j<fin->VS.size(); j++) {
            xin = &(fin->VS[j].x);
            yin = &(fin->VS[j].y);
            fout->VS[j].x.clear();
            fout->VS[j].y.clear();
            for( uint i=0; i<yin->size()-1; ++i ){
                dx = (*xin)[i+1] - (*xin)[i];
                fout->VS[j].push_xy(
                    ((*xin)[i+1] + (*xin)[i] ) / 2,
                    dx==0 ? 0 : ((*yin)[i+1] - (*yin)[i]) / dx );
            }
        }
        if(!(NOlm::overwrite)) NOlm::OloAdd(&ftmp);
    }
}

void NOperate::FunctionalAutocorrelate()
{
    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *fin, ftmp, *fout;
    while((fin=fiter())) {
        OUT_TO_IN( NOlm::overwrite, fin, ftmp, fout );
        fout->yco.name = "ac{" + fin->yco.name + "}";
        fout->yco.unit = fin->yco.unit + "^2";
        fout->lDoc.push_back( "ofac" );
        vector<double> *yin;
        for (uint j=0; j<fin->VS.size(); j++) {
            //xin = &(fin->VS[j].x);
            yin = &(fin->VS[j].y);
            uint n = yin->size();
            // cout << "-> -> we have " << n << "data points\n";
            if (n<=2)
                throw string ("AC n<=2");
            throw string("AC ausser Betrieb");
/*
  double dx = ((*xin)[n-1] - (*xin)[0]) / n;
  // cout << "-> -> step width is " << dx << "\n";
  if (dx<=0) {
  cout << "AC dx<=0\n";
  return 2;
  }
  double ym = 0, ys = 0;
  int i, j;
  // calculate mean:
  for (i=0; i<n; i++) {
  ym += (*yin)[i];
  ys += SQR((*yin)[i]);
  }
  ym /= n;
  ys = ys/n - SQR(ym);;
  if (normed) 
  cout << "      <y^2> = " << ys << "\n";
  int nout = n/2;
  vector<double> xtmp(nout); 
  vector<double> ytmp(nout);
  // calculate autocorrelation:
  for (i=0; i<nout; i++) {
  xtmp[i] = i*dx;
  double ya = 0;
  for (j=0; j+i<n; j++) {
  ya += ((*yin)[j]-ym) * ((*yin)[j+i]-ym);
  }
  ytmp[i] = ya / j;
  if (normed) ytmp[i] /= ys;
  }
  fout->VS[j].x = xtmp;
  fout->VS[j].y = ytmp;
*/
        }
        if(!(NOlm::overwrite)) NOlm::OloAdd(&ftmp);
    }
}

//***************************************************************************//
//* old-style integral operations                                           *//
//***************************************************************************//

namespace NIntOld {
    static int mod;
    double Eval(CScan S);
    static double arg2;
    static int iarg2;
};

void NIntOld::Opr(void) 
{
    printf("integral properties:\n"
           "   1:                     2: one y\n"
           "   3: <y> average         4: stddev in y\n"
           "   5: x(max y)            6: y(max y)\n"
           "   7: x(min y)            8: y(min y)\n"
           "   9: sum                10: integral\n"
           "  11: x(int<arg)\n"
           "  13: max x(y>0)\n"
           "  15: center of gravity  16: stddev in x\n"
           "  17: correl.coeff (x<->y)\n"
        );
    mod = iask("Choice", mod );
	
    if        (mod== 0) {
        return;
    } else if (mod== 2) {
        iarg2 = iask("channel number", iarg2);
    } else if (mod== 3) {
    } else if (mod== 4) {
    } else if (mod== 5) {
    } else if (mod== 6) {
    } else if (mod== 7) {
    } else if (mod== 8) {
    } else if (mod== 9) {
    } else if (mod==10) {
    } else if (mod==11) {
        arg2 = dask("second argument", arg2);
    } else if (mod==13) {
    } else if (mod==15) {
    } else if (mod==16) {
    } else if (mod==17) {
    } else
        throw string( "invalid option" );
	
    // CList FOpr = NOlm::FSel;
    NOlm::IterateD fiter;
    if (!NOlm::overwrite) NOlm::SelNew();
    COld *fin, fout;
    CScan S;
    CCoord zco;
    double xval, zval, result;
    bool savable;
    int nz;
    while( (fin=fiter()) ) {
        nz = fin->ZCo.size();
        savable = nz > 0;

        fout = *fin;
        fout.VS.clear();
        if (savable) {
            zco = fin->ZCo.back();
            fout.ZCo.pop_back();
            fout.xco = zco;
        }
        if        (mod== 2) {
            fout.yco.name = fin->yco.name + " (#"
                + strg(iarg2) + ")";
        } else if (mod== 3) {
            fout.yco.name = "<" + fin->yco.name + ">";
            fout.yco.unit = fin->yco.unit;
        } else if (mod== 4) {
            fout.yco.name = "sigma(" + fin->yco.name + ")";
            fout.yco.unit = "" ;
        } else if (mod== 5) {
            fout.yco.name = fin->xco.name + "_max";
            fout.yco.unit = fin->xco.unit;
        } else if (mod== 6) {
            fout.yco.name = fin->yco.name + "_max";
            fout.yco.unit = fin->yco.unit;
        } else if (mod== 7) {
            fout.yco.name = fin->xco.name + "_min";
            fout.yco.unit = fin->xco.unit;
        } else if (mod== 8) {
            fout.yco.name = fin->yco.name + "_min";
            fout.yco.unit = fin->yco.unit;
        } else if (mod== 9) {
            fout.yco.name = "sum " + fin->yco.name;
            fout.yco.unit = fin->yco.unit;
        } else if (mod==10) {
            fout.yco.name = 
                "I d" + fin->xco.name + " " + fin->yco.name;
            fout.yco.unit = fin->yco.unit + " * " + fin->xco.unit;
        } else if (mod==11) {
            fout.yco.name = fin->xco.name + 
                string("[integral{") + 
                fin->yco.name + "} < " + strg(arg2) + "]";
            fout.yco.unit = fin->xco.unit;
        } else if (mod==13) {
            fout.yco.name = "max " + fin->xco.name +
                " with positive " + fin->yco.name;
            fout.yco.unit = fin->xco.unit;
        } else if (mod==15) {
            fout.yco.name = fin->xco.name + 
                "[center of gravity of " + 
                fin->yco.name + "]";
            fout.yco.unit = fin->xco.unit;
        } else if (mod==16) {
            fout.yco.name = "sigma(" + fin->xco.name + ")";
            fout.yco.unit = "" ;
        } else if (mod==17) {
            fout.yco.name = "rho(" + fin->xco.name +
                "|" + fin->yco.name + ")";
            fout.yco.unit = "" ;
        } else {
            cout << "PROGRAM ERROR not implemented\n";
            return;
        }
        fout.lDoc.push_back( "oio "+strg(mod)+" # y = "+fout.yco.name );

        S.Clear();
        S.z = fin->VS[0].z;
        S.z.pop_back();
        for (uint j=0; j<fin->nScan(); j++) {
            result = Eval(fin->VS[j]);
            if (!savable) {
                cout << "results: " << result << "\n";
                break;
            }
            xval = fin->VS[j].z[nz-1];
            S.push_xy(xval, result);
            if (nz>=2) { // new scan if jump in other z values
                zval = fin->VS[j].z[nz-2];
                if ((j+1)<fin->VS.size() && 
                    fin->VS[j+1].z[nz-2]!=zval) {
                    fout.VS.push_back(S);
                    S.Clear();
                    S.z = fin->VS[j+1].z;
                    S.z.pop_back();
                }
            }
        }
        if (savable) {
            fout.VS.push_back(S);
            NOlm::OloAdd(&fout);
        }
    }
};

double NIntOld::Eval(CScan S)
{
    int i, i0;
    double r0, r1, r2, s0, s1, v, dr, area, triangle, p, center;
    int n = S.size();
    switch(mod) {
    case  2:
        if (iarg2<0 || iarg2>=n) 
            throw string( "i out of range" );
        return S.y[iarg2];
    case  3:
        r0 = 0;
        for (i=0; i<n; i++) {
            r0 += S.y[i];
        }
        return r0/n;
    case 4:
        r0 = 0; s0 = 0;
        for (i=0; i<n; i++) {
            r0 += S.y[i] / n;
        }
        for (i=0; i<n; i++) {
            s0 += SQR(S.y[i]-r0) / n;
        }
        return sqrt(s0);
    case  5:
        i0 = 0;
        for (i=1; i<n; i++) {
            if( S.y[i]>S.y[i0] ) i0 = i;
        }
        return S.x[i0];
    case  6:
        i0 = 0;
        for (i=1; i<n; i++) {
            if( S.y[i]>S.y[i0] ) i0 = i;
        }
        return S.y[i0];
    case  7:
        i0 = 0;
        for (i=1; i<n; i++) {
            if( S.y[i]<S.y[i0] ) i0 = i;
        }
        return S.x[i0];
    case  8:
        i0 = 0;
        for (i=1; i<n; i++) {
            if( S.y[i]<S.y[i0] ) i0 = i;
        }
        return S.y[i0];
    case  9: // sum_i y
        r0 = 0;
        for (i=0; i<n; i++)
            r0 += S.y[i];
        return r0;
    case 10: // integral dx y
        r0 = 0;
        for (i=0; i<n-1; i++) {
            if (S.x[i+1]<=S.x[i])
                throw string( "x not sorted" );
            r0 += (S.x[i+1]-S.x[i])*(S.y[i+1]+S.y[i])/2;
        }
        return r0;
    case 11:
        r0 = 0; r1 = 0;
        for (i=0; i<n-1; i++) {
            if (S.x[i+1]<=S.x[i])
                throw string( "x not sorted" );
            r0 += (S.x[i+1]-S.x[i])*(S.y[i+1]+S.y[i])/2;
        }
        if (r0<=0)
            throw string( "int <= 0" );
        for (i=0; i<n-1; i++) {
            dr = (S.x[i+1]-S.x[i])*(S.y[i+1]+S.y[i])/2;
            if ((r1+dr)/r0>=arg2) break;
            r1 += dr;
        }
        area = arg2*r0 - r1;
        triangle = (S.y[i+1]-S.y[i])/(S.x[i+1]-S.x[i])/2;
        if (triangle==0) {
            return S.x[i]+area/dr;
        } else {
            p = S.y[i]/2/triangle;
            return S.x[i] - p + 
                (triangle>0?+1:-1) * sqrt(p*p + area/triangle);
        }
    case 13:
        for (i=n-1; i>=0; i--)
            if (S.y[i]>0) break;
        return S.x[i];
    case 15: // center of gravity
        r0 = 0; r1 = 0;
        for (i=0; i<n; i++) {
            r0 += S.y[i];
            r1 += S.y[i] * S.x[i];
        }
        return r0>0 ? r1/r0 : 0;
    case 16: // stddev in x
        r0 = 0; r1 = 0; r2 = 0;
        for (i=0; i<n; i++) {
            r0 += S.y[i];
            r1 += S.y[i] * S.x[i];
        }
        if( r0<=0 )
            throw string( "r0 <= 0" );
        center = r1/r0;
        for (i=0; i<n; i++) {
            r2 += S.y[i] * SQR(S.x[i]-center);
        }
        return sqrt( r2/r0 );
    case 17: // correlation coefficient
        r0 = 0; r1 = 0; s0 = 0; s1 = 0; v = 0;
        for (i=0; i<n; i++) {
            r0 += S.x[i] / n;
            r1 += S.y[i] / n;
        }
        for (i=0; i<n; i++) {
            s0 += SQR(S.x[i]-r0) / n;
            s1 += SQR(S.y[i]-r1) / n;
            v  += (S.x[i]-r0) * (S.y[i]-r1) / n;
        }
        return v / sqrt(s0*s1);
    default:
        throw string( "PROGRAM ERROR Integral not implemented" ); 
    }
	
}
